#include <stdlib.h>
#include <iostream>
#include <ctime>
#include <cmath>
#include <cv.h>
#include <highgui.h>

#include "../lib/orga2.h" 
//The OpenCL C++ bindings, with exceptions
#define __CL_ENABLE_EXCEPTIONS
#define WORK_GROUP_SIZE_X 16
#define WORK_GROUP_SIZE_Y 16
#define KERFILE "Sharpen.cl" //source of the kernels
#define KERFUNC "sharpen" //name of the kernel to be used
#define NCHANNELS 3
#define BLOCK_SIZE 16	//el ancho o alto del bloque
#include <CL/cl.hpp>


using namespace std;

void cvPrint(char* str, IplImage* img){
	CvFont font;
	double hScale=1.0;
	double vScale=1.0;
	int    lineWidth=2;
	cvInitFont(&font,CV_FONT_HERSHEY_SIMPLEX|CV_FONT_ITALIC, hScale,vScale,0,lineWidth);
	cvPutText (img,str,cvPoint(40,40), &font, cvScalar(255,255,0));
}

int getMinNumOfBlocksRequired(int dim){
	return dim / BLOCK_SIZE + (dim % BLOCK_SIZE == 0 ? 0 : 1);
}

int main( int argc, char** argv )
{
	try{
		time_t start,end;
		IplImage  *frame, *outFrame;
		int key, height, width, step, counter;
		if(argc < 2){
			cout << "Usage: ./VideoConvolution <raw-video>" << endl;
			exit(1);
		}

		/* load the AVI file */
		CvCapture *capture = cvCaptureFromAVI( argv[1] );

		/* always check */
		if( !capture ){
			cout << "Error: failed to load video file" << endl;
			return 1;    
		}

		/* get fps, needed to set the delay */
		int fps = ( int )cvGetCaptureProperty( capture, CV_CAP_PROP_FPS );

		/* display video */
		cvNamedWindow( "inputVideo", CV_WINDOW_AUTOSIZE );
		cvMoveWindow( "inputVideo", 10, 10);

		cvNamedWindow( "outputVideo", CV_WINDOW_AUTOSIZE );
		cvMoveWindow( "outputVideo", 1000, 10);

		vector<cl::Platform> platforms;
		cl::Platform::get(&platforms);
		vector<cl::Device> devices;
		platforms.front().getDevices(CL_DEVICE_TYPE_GPU, &devices);
		cl::Context context(devices);
		cl::Device device = devices.front();

		//load kernel from filename
		std::string kernelSource = convertToString(KERFILE);
		cl::Program::Sources sourceCode;
		sourceCode.push_back(std::make_pair(kernelSource.c_str(), kernelSource.size()));
		
		//create a program
		cl::Program program = cl::Program(context, sourceCode);
		
		//build it for all devices
		try{ program.build(devices); }
		catch(cl::Error &err){
			std::cerr << "Error building kernel " << KERFILE << std::endl;
			tr(devices, it){
				std::cerr << "\t" << it->getInfo<CL_DEVICE_NAME>() << " log: " 
				<< program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(*it) << std::endl;
			}
			return -1;
		}
		cl::CommandQueue cmdQueue(context, device);

		//datos para computar el tamanio del video:
		height = (int) cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_HEIGHT);
		width  = (int) cvGetCaptureProperty(capture, CV_CAP_PROP_FRAME_WIDTH);
		cout << width << " x " << height << " --> ";
		//tamanio del frame = height*width*channels*tam del dato
		cl::Buffer inputBuffer(context, CL_MEM_READ_WRITE, height*width*NCHANNELS*sizeof(uchar));	
		cl::Buffer outputBuffer(context, CL_MEM_READ_WRITE, height*width*NCHANNELS*sizeof(uchar));
		//initialize kernel
		cl::Kernel kernel(program, KERFUNC);
		cl::KernelFunctor kernel_func = kernel.bind(cmdQueue, 
			cl::NDRange(BLOCK_SIZE * getMinNumOfBlocksRequired(width) ,BLOCK_SIZE * getMinNumOfBlocksRequired( height)), 
			cl::NDRange(BLOCK_SIZE,BLOCK_SIZE));

		cout << BLOCK_SIZE *  getMinNumOfBlocksRequired(width) << " x "<<BLOCK_SIZE *  getMinNumOfBlocksRequired( height) << endl;;
		time(&start);
		counter = 0;

		outFrame = cvCreateImage(cvSize(width,height), IPL_DEPTH_8U, NCHANNELS);
		while( key != 1048689){ //'q' ) {
			frame = cvQueryFrame( capture );
			if( !frame ) break;
			step = frame->widthStep;

			cmdQueue.enqueueWriteBuffer(inputBuffer, true, 0, height*width*NCHANNELS*sizeof(uchar), frame->imageData);	//write buf 2 device main mem

			//execute
			kernel_func(inputBuffer, outputBuffer, width, height, step);

			cmdQueue.enqueueReadBuffer(outputBuffer, true, 0, height*width*NCHANNELS*sizeof(uchar), outFrame->imageData);	//read to host mem from dev mem
			
			//outFrame = convolve_copy(frame, m);
			
//			cvShowImage( "inputVideo", frame );
			cvShowImage( "outputVideo", outFrame );
			time(&end);
			++counter;
			double sec = difftime(end,start);
			double fpsRatio = counter/sec;
			cout << fpsRatio << "\t fps " << endl;
			/* quit if user press 'q' */
			key = cvWaitKey( 1000 / fps);
		}

		/* free memory */
		cvReleaseCapture( &capture );
		cvDestroyWindow( "video" );
	} catch (cl::Error& err){
	  clErr(err);
	  return -1;
	}

	return 0;
}
